package fr.contralia.client.example;

import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.nio.charset.Charset;
import java.util.Properties;

import org.apache.http.HttpEntity;
import org.apache.http.HttpHost;
import org.apache.http.HttpResponse;
import org.apache.http.StatusLine;
import org.apache.http.auth.AuthScope;
import org.apache.http.auth.UsernamePasswordCredentials;
import org.apache.http.client.AuthCache;
import org.apache.http.client.ResponseHandler;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.HttpRequestBase;
import org.apache.http.client.protocol.ClientContext;
import org.apache.http.conn.params.ConnRoutePNames;
import org.apache.http.entity.mime.MultipartEntity;
import org.apache.http.entity.mime.content.FileBody;
import org.apache.http.entity.mime.content.StringBody;
import org.apache.http.impl.auth.BasicScheme;
import org.apache.http.impl.client.BasicAuthCache;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.protocol.BasicHttpContext;
import org.apache.http.util.EntityUtils;

/**
 * Util class implementing Contralia Core API
 *
 * @author lfaizant
 *
 */
public class Common {

	private static final Charset charset = Charset.forName("UTF-8");

	/**
	 * HTTP methods
	 */
	public enum HttpMethod {
		GET,
		POST
	}

	/**
	 * Class that reads and holds the properties (url, login, password, proxy, etc.)
	 */
	static class Props {
		URL url;
		String login;
		String password;
		HttpHost proxy;
		String offerCode;
		String organizationalUnitCode;

		static Props props = null;

		synchronized static Props getInstance() throws IOException {
			if (props == null) {
				props = new Props();
			}
			return props;
		}

		Props() throws IOException {

			Properties props;
			String userHome = System.getProperty("user.home");
			File userPropsFile = new File(userHome, "contralia-client-example/config.properties");
			if (userPropsFile.exists()) {
				// Read the user's custom prop file
				props = readProperties(userPropsFile.getAbsolutePath());
			} else {
				throw new IOException(userPropsFile.getAbsolutePath() + " file not found, please create it using template file " + new File( "conf/config.properties").getAbsolutePath());
			}

			url = new URL(props.getProperty("url"));
			login = props.getProperty("login");
			password = props.getProperty("password");
			if (props.containsKey("proxy.host") && props.getProperty("proxy.host").length() != 0) {
				proxy = new HttpHost(props.getProperty("proxy.host"), Integer.valueOf(props.getProperty("proxy.port")));
			}
			offerCode = props.getProperty("offerCode");
			organizationalUnitCode = props.getProperty("organizationalUnitCode");
		}
	}

	/**
	 * Exception raised in case of server response status is not 200.
	 */
	static public class HttpException extends IOException {
		public int statusCode;
		public String reasonPhrase;
		public String response;

		HttpException(int statusCode, String reasonPhrase, String response) {
			this.statusCode = statusCode;
			this.reasonPhrase = reasonPhrase;
			this.response = response;
		}
	}

	/**
	 * Our Custom response handler. In case of error, this handler extracts the XML response if any and returns it
	 * in a HttpException, with the status code and reason text.
	 */
	static class CustomResponseHandler implements ResponseHandler<String> {

		CustomResponseHandler() {
		}

		public String handleResponse(HttpResponse response) throws IOException {
			StatusLine statusLine = response.getStatusLine();
			HttpEntity entity = response.getEntity();
			if (statusLine.getStatusCode() != 200) {
				String responseAsString = entity == null ? null : EntityUtils.toString(entity);
				EntityUtils.consume(entity);
				throw new HttpException(statusLine.getStatusCode(), statusLine.getReasonPhrase(), responseAsString);
			} else {
				return entity == null ? null : EntityUtils.toString(entity);
			}
		}
	}

	/**
	 * Reads a properties file
	 * 
	 * @param path
	 * 		Properties file location
	 * @return
	 * 		Properties stored in the file
	 * @throws IOException
	 */
	private static Properties readProperties(String path) throws IOException {
		try {
			FileInputStream input = new FileInputStream(new File(path));
			Properties properties = new Properties();
			properties.load(input);
			input.close();

			return properties;
		}
		catch (IOException e) {
			throw new IOException("Error while reading properties file", e);
		}
	}
	
	/**
	 * Reads text file content
	 * 
	 * @param path
	 * 		Text file location
	 * @return
	 * 		Text file content
	 * @throws IOException
	 */
	private static String readFile(String path) throws IOException {
		try {
			File file = new File(path);
			DataInputStream input = new DataInputStream(new FileInputStream(file));
			byte[] bytes = new byte[(int) file.length()];
			input.readFully(bytes);
			
			input.close();
			
			return new String(bytes);
		}
		catch (IOException e) {
			throw new IOException("Error while reading file", e);
		}
	}

	/**
	 * Creates a HttpClient.
	 * IMPORTANT : creating a HttpClient for each request is not optimized at all. For better performances consider reusing a static HttpClient and configure it with a connection pool.
	 */
	private static DefaultHttpClient buildHttpClient(Props props) throws IOException {
		// Init HTTP client with Contralia URL, login and password
		DefaultHttpClient httpClient = new DefaultHttpClient();
		httpClient.getCredentialsProvider().setCredentials(
				new AuthScope(props.url.getHost(), props.url.getPort()),
				new UsernamePasswordCredentials(props.login, props.password));
		// Set proxy if any
		if (props.proxy != null) {
			httpClient.getParams().setParameter(ConnRoutePNames.DEFAULT_PROXY, props.proxy);
		}

		return httpClient;
	}

	/**
	 * Creates a context to pass with each request.
	 */
	private static BasicHttpContext buildContext(Props props) {
		// Set HTTP context
		BasicHttpContext context = new BasicHttpContext();
		AuthCache cache = new BasicAuthCache();
		BasicScheme scheme = new BasicScheme();
		cache.put(new HttpHost(props.url.getHost(), props.url.getPort(), props.url.getProtocol()), scheme);
		context.setAttribute(ClientContext.AUTH_CACHE, cache);

		return context;
	}

	/**
	 * Calls a Contralia API and returns result as string
	 *
	 * @param api
	 * 		API URL
	 * @param params
	 * 		Call parameters
	 * @param method
	 * 		HTTP method to use
	 * @return
	 * 		HTTP request result as string
	 * @throws IOException
	 */
	private static String getResponseAsString(String api, MultipartEntity params, HttpMethod method) throws IOException {
		Props props = Props.getInstance();
		DefaultHttpClient httpClient = buildHttpClient(props);
		BasicHttpContext context = buildContext(props);

		// Set HTTP request according to the HTTP method to use
		String response;
		if (method == HttpMethod.GET) {
			HttpGet request = new HttpGet(props.url.toString() + api);
			// Execute HTTP request
			response = httpClient.execute(request, new CustomResponseHandler(), context);
			// Close HTTP request
			request.reset();
		}
		else {
			HttpPost request = new HttpPost(props.url.toString() + api);
			// Set HTTP request parameters
			request.setEntity(params);
			// Execute HTTP request
			response = httpClient.execute(request, new CustomResponseHandler(), context);
			// Close HTTP request
			request.reset();
		}

		// Close HTTP connection
		httpClient.getConnectionManager().shutdown();
		return response;
	}
	/**
	 * Calls a Contralia API and returns result as file
	 * 
	 * @param api
	 * 		API URL
	 * @param params
	 * 		Call parameters
	 * @param method
	 * 		HTTP method to use
	 * @param outputFile
	 * 		File where to write the final signed doc
	 * @return
	 * 		Absolute path to the written file
	 * @throws IOException
	 */
	private static void getResponseAsFile(String api, MultipartEntity params, HttpMethod method, File outputFile) throws IOException {
		Props props = Props.getInstance();
		DefaultHttpClient httpClient = buildHttpClient(props);
		BasicHttpContext context = buildContext(props);

		// Set HTTP request according to the HTTP method to use
		HttpRequestBase request;
		if (method == HttpMethod.GET)
			request = new HttpGet(props.url.toString() + api);
		else {
			request = new HttpPost(props.url.toString() + api);
			((HttpPost)request).setEntity(params);
		}
		// Execute HTTP request
		HttpEntity response = httpClient.execute(request, context).getEntity();

		// Init output stream
		if (outputFile.exists()) {
			outputFile.delete();
		}
		outputFile.createNewFile();
		FileOutputStream output = new FileOutputStream(outputFile);
		byte[] buffer = new byte[1024];
		int length;

		// Write HTTP request result in file
		InputStream input = response.getContent();
		while((length = input.read(buffer)) != -1)
			output.write(buffer, 0, length);

		// Close input and output streams
		output.flush();
		output.close();
		input.close();

		// Close HTTP request
		request.reset();
		// Close HTTP connection
		httpClient.getConnectionManager().shutdown();
	}
	
	/**
	 * Calls 'transaction/initiate' Contralia API to initiate a transaction
	 *
	 * @param customRef
	 * @param signatoriesCount
	 * @param testMode
	 * @return
	 * @throws IOException
	 */
	public static String initiate(String customRef, int signatoriesCount, boolean testMode) throws IOException {
		try {
			Props props = Props.getInstance();

			// Set HTTP request parameters
			MultipartEntity params = new MultipartEntity();
			params.addPart("offerCode", new StringBody(props.offerCode));
			params.addPart("organizationalUnitCode", new StringBody(props.organizationalUnitCode));
			params.addPart("customRef", new StringBody(customRef, charset));
			
			if (signatoriesCount != 1)
				params.addPart("signatoriesCount", new StringBody(new Integer(signatoriesCount).toString()));
			if (testMode)
				params.addPart("testMode", new StringBody("true"));
			
			// Call the API
			String result = getResponseAsString("/api/v2/" + props.offerCode + "/transactions", params, HttpMethod.POST);
			return result;
		}
		catch (IOException e) {
			throw new IOException("Error while initializing transaction", e);
		}
	}
	
	/**
	 * Calls 'transaction/upload' Contralia API to set document to sign
	 *
	 * @param id
	 * @param file
	 * @param name
	 * @return
	 * @throws IOException
	 */ 
	public static String upload(String id, String file, String name) throws IOException {
		try {
			// Set HTTP request parameters
			MultipartEntity params = new MultipartEntity();
			params.addPart("file", new FileBody(new File(file)));
			if (name != null)
				params.addPart("name", new StringBody(name, charset));

			// Call the API
			String result = getResponseAsString("/api/v2/transactions/" + id + "/document", params, HttpMethod.POST);
			return result;
		}
		catch (IOException e) {
			throw new IOException("Error while uploading document", e);
		}
	}
	
	/**
	 * Calls 'transaction/signatory' Contralia API to set signatory
	 * 
	 * @param id
	 * @param position
	 * @param fieldNumber
	 * @param civility
	 * @param lastname
	 * @param firstname
	 * @param email
	 * @param phone
	 * @param street
	 * @param complements
	 * @param postalCode
	 * @param city
	 * @param country
	 * @param companyName
	 * @param role
	 * @param companyRegistrationNumber
	 * @param signatureType
	 * @return
	 * @throws IOException
	 */
	public static String signatory(String id, String position, String fieldNumber, String civility, String lastname, String firstname, String email,
								   String phone, String street, String complements, String postalCode, String city, String country, String companyName,
								   String role, String companyRegistrationNumber, String signatureType) throws IOException {
		try {
			// Set HTTP request parameters
			MultipartEntity params = new MultipartEntity();
			
			if (position != null)
				params.addPart("position", new StringBody(position));
			if (fieldNumber != null)
				params.addPart("fieldNumber", new StringBody(fieldNumber));
			if (civility != null)
				params.addPart("civility", new StringBody(civility));
			params.addPart("lastname", new StringBody(lastname, charset));
			params.addPart("firstname", new StringBody(firstname, charset));
			if (email != null)
				params.addPart("email", new StringBody(email, charset));
			params.addPart("phone", new StringBody(phone));
			params.addPart("adress.street", new StringBody(street, charset));
			if (complements != null)
				params.addPart("adress.complements", new StringBody(complements, charset));
			params.addPart("adress.postalCode", new StringBody(postalCode, charset));
			params.addPart("adress.city", new StringBody(city, charset));
			if (country != null)
				params.addPart("adress.country", new StringBody(country, charset));
			if (companyName != null)
				params.addPart("companyName", new StringBody(companyName, charset));
			if (role != null)
				params.addPart("role", new StringBody(role, charset));
			if (companyRegistrationNumber != null)
				params.addPart("companyRegistrationNumber", new StringBody(companyRegistrationNumber));
			if (signatureType != null)
				params.addPart("signatureType", new StringBody(signatureType));
			
			// Call the API
			String result = getResponseAsString("/api/v2/transactions/" + id + "/signatory", params, HttpMethod.POST);
			return result;
		}
		catch (IOException e) {
			throw new IOException("Error while setting signatory", e);
		}
	}
	
	/**
	 * Calls 'transaction/pad' Contralia API to sign document
	 *
	 * @param id
	 * @param biometryFile
	 * @return
	 * @throws IOException
	 */
	public static String pad(String id, String biometryFile) throws IOException {
		try {
			// Set HTTP request parameters
			String biometry = readFile(biometryFile);
			MultipartEntity params = new MultipartEntity();
			params.addPart("biometry", new StringBody(biometry));
			
			// Call the API
			String result = getResponseAsString("/api/v2/signatures/" + id + "/pad", params, HttpMethod.POST);
			return result;
		}
		catch (IOException e) {
			throw new IOException("Error while signing document", e);
		}
	}
	
	/**
	 * Calls 'transaction/terminate' Contralia API to close transaction
	 *
	 * @param id
	 * @return
	 * @throws IOException
	 */
	public static String terminate(String id) throws IOException {
		try {
			// Call the API
			String result = getResponseAsString("/api/v2/transactions/" + id + "/terminate", null, HttpMethod.POST);
			return result;
		}
		catch (IOException e) {
			throw new IOException("Error while terminating transaction", e);
		}
	}
	
	/**
	 * Calls 'transaction/finalDoc' Contralia API to retrieve signed document
	 * 
	 * @param id
	 * @return
	 * @throws IOException
	 */
	public static void finalDoc(String id, String name, File outputFile) throws IOException {
		try {
			// Call the API
			String url = "/api/v2/transactions/" + id + "/finalDoc";
			if (name != null)
				url += "?name=" + name;
			getResponseAsFile(url, null, HttpMethod.GET, outputFile);
		}
		catch (IOException e) {
			throw new IOException("Error while retrieving document", e);
		}
	}

	/**
	 * Calls 'transaction/signHashes' Contralia API to sign a hash
	 *
	 * @param hash
	 * @return
	 * @throws IOException
	 */
	public static String signHash(String transactionId, String signatureId, String hash) throws IOException {
		try {
			// Call the API
			MultipartEntity params = new MultipartEntity();
			params.addPart("signatureId", new StringBody(signatureId));
			params.addPart("hashes", new StringBody(hash));
			params.addPart("hashAlgo", new StringBody("SHA-256"));
			params.addPart("signatureStandard ", new StringBody("PKCS7"));
			String result = getResponseAsString("/api/v2/transactions/" + transactionId + "/signHashes", params, HttpMethod.POST);
			return result;
		}
		catch (IOException e) {
			throw new IOException("Error while signing hash", e);
		}
	}

	/**
	 * Calls 'transaction/timestampHashes' Contralia API to timestamp a hash
	 *
	 * @param hash
	 * @return
	 * @throws IOException
	 */
	public static String timestampHash(String transactionId, String hash) throws IOException {
		try {
			// Call the API
			MultipartEntity params = new MultipartEntity();
			params.addPart("hashes", new StringBody(hash));
			params.addPart("hashAlgo", new StringBody("SHA-256"));
			String result = getResponseAsString("/api/v2/transactions/" + transactionId + "/timestampHashes", params, HttpMethod.POST);
			return result;
		}
		catch (IOException e) {
			throw new IOException("Error while timestamping hash", e);
		}
	}

}
