package fr.contralia.client.example;

import com.itextpdf.text.Font;
import com.itextpdf.text.Image;
import com.itextpdf.text.pdf.*;
import com.itextpdf.text.pdf.security.*;
import org.apache.commons.codec.binary.Hex;
import org.bouncycastle.jce.provider.BouncyCastleProvider;

import java.io.*;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.nio.file.StandardCopyOption;
import java.security.GeneralSecurityException;
import java.security.MessageDigest;
import java.security.Security;
import java.util.Base64;
import java.util.Calendar;
import java.util.List;
import java.util.concurrent.atomic.AtomicInteger;

// Class that shows how to sign a PDF with the Transaction/signHashes API.
// WARNING : this class is not thread safe : create a new instance for each signature
public class PdfSigner {

	// WARNING : estimatedSignatureSize must be a static variable that is reused for all signatures. It contains the estimated size of the PKCS7 signature.
	// Unfortunately, it is not possible to know this size in advanced and iText needs it to
	// The principle is :
	// - try initial estimated size
	// - in case of error, increase the estimatedSize
	// - reuse new estimatedSize to avoid error at every signature
	// Once estimatedSize is correct, it must be reused.
	private static final AtomicInteger estimatedSignatureSize = new AtomicInteger(5000);
	// Same logic for estimatedTimestampSize
	private static final AtomicInteger estimatedTimestampSize = new AtomicInteger(100);

	private File pdfOutput;
	private File tmpInput;

	private void copyOutputToTmp() throws Exception {
		Files.copy(Paths.get(pdfOutput.getAbsolutePath()), Paths.get(tmpInput.getAbsolutePath()), StandardCopyOption.REPLACE_EXISTING);
	}


	public static class SignParams {
		public String transactionId;
		public String signatureId;
		public String fieldName;
		public Image fieldImage;
		public String reason;
		public String location;
		public String contact;
		public String creator;
	}

	private static String pdfFieldsV1_SIG_CONTRACTANT = "SignatureContractant";
	private static String pdfFieldsV1_SIG_FOURNISSEUR = "SignatureFournisseur";
	private static String pdfFieldsV1_SIG_CONTRALIA = "TamponContralia";

	private static String pdfFieldsV2_SIG_CONTRACTANT = "SignatorySignature";
	private static String pdfFieldsV2_SIG_FOURNISSEUR = "OrganizationSignature";
	private static String pdfFieldsV2_SIG_CONTRALIA = "SealingSignature";

	private static void assertTrue(boolean condition, String message) throws Exception {
		if (!condition) {
			throw new Exception(message);
		}
	}

	// WARNING : this class is not thread safe : create a new instance for each signature
	static class MyExternalSignatureContainer implements ExternalSignatureContainer {
		String transactionId;
		String signatureId;
		byte[] pkcs7Signature;
		PdfDictionary sigDic;

		MyExternalSignatureContainer(String transactionId, String signatureId) {
			this.transactionId = transactionId;
			this.signatureId = signatureId;

			sigDic = new PdfDictionary();
			sigDic.put(PdfName.FILTER, PdfName.ADOBE_PPKLITE);
			sigDic.put(PdfName.SUBFILTER, PdfName.ETSI_CADES_DETACHED);
		}

		@Override
		public void modifySigningDictionary(PdfDictionary dic) {
			dic.putAll(sigDic);
		}

		@Override
		public byte[] sign(InputStream input) throws GeneralSecurityException {
			try {
				byte[] hashedBytes = Hash.hash(input, "SHA-256");
				String hash = Hex.encodeHexString(hashedBytes);
				String xmlResult = Common.signHash(transactionId, signatureId, hash);
				// Extract signature from XMl
				XmlReader.Node2 xml = XmlReader.parse(xmlResult);
				assertTrue(xml.getChild("signature").getAttribute("hash").equals(hash), "Incorrect signature hash");
				String base64Pkcs7Signature = xml.getChild("signature").getText();
				pkcs7Signature = Base64.getDecoder().decode(base64Pkcs7Signature);
				return pkcs7Signature;

			} catch(Exception e) {
				throw new GeneralSecurityException("Error signing hash", e);
			}
		}
	}

	// WARNING : this class is not thread safe : create a new instance for each signature
	static class MyTSAClient implements TSAClient {

		String transactionId;
		byte[] timestamp;

		MyTSAClient(String transactionId) {
			this.transactionId = transactionId;
		}

		@Override
		public MessageDigest getMessageDigest() throws GeneralSecurityException {
			return MessageDigest.getInstance("SHA-256");
		}

		@Override
		public byte[] getTimeStampToken(byte[] hashBytes) throws Exception {
			try {
				String hash = Hex.encodeHexString(hashBytes);
				String xmlResult = Common.timestampHash(transactionId, hash);
				// Extract timestamp from XML
				XmlReader.Node2 xml = XmlReader.parse(xmlResult);
				assertTrue(xml.getChild("timestamp").getAttribute("hash").equals(hash), "Incorrect timestamp hash");
				String base64Timestamp = xml.getChild("timestamp").getText();
				timestamp = Base64.getDecoder().decode(base64Timestamp);
				return timestamp;

			} catch(Exception e) {
				throw new GeneralSecurityException("Error timestamping hash", e);
			}
		}

		public int getTokenSizeEstimate() {
			return estimatedTimestampSize.get();
		}
	}


	public PdfSigner(File pdfInput, File pdfOutput, File tmpInput) throws Exception {
		this.pdfOutput = pdfOutput;
		this.tmpInput = tmpInput;

		// Make tmp file our working file to not modify the input file
		Files.copy(Paths.get(pdfInput.getAbsolutePath()), Paths.get(tmpInput.getAbsolutePath()), StandardCopyOption.REPLACE_EXISTING);
	}

	// Replace text fields managed by Contralia by signature fields
	public void replaceFields() throws Exception {
		PdfReader reader = new PdfReader(new FileInputStream(tmpInput));
		PdfStamper stamper = new PdfStamper(reader, new FileOutputStream(pdfOutput), '\0', false);

		for (String fieldName : reader.getAcroFields().getFields().keySet()) {
			if (fieldName.startsWith(pdfFieldsV1_SIG_CONTRACTANT) || fieldName.startsWith(pdfFieldsV2_SIG_CONTRACTANT) ||
				fieldName.startsWith(pdfFieldsV1_SIG_FOURNISSEUR) || fieldName.startsWith(pdfFieldsV2_SIG_FOURNISSEUR) ||
				fieldName.startsWith(pdfFieldsV1_SIG_CONTRALIA) || fieldName.startsWith(pdfFieldsV2_SIG_CONTRALIA))
			{
				if (reader.getAcroFields().getFieldType(fieldName) == AcroFields.FIELD_TYPE_TEXT) {
					AcroFields.FieldPosition pos = reader.getAcroFields().getFieldPositions(fieldName).get(0);
					stamper.getAcroFields().removeField(fieldName);
					PdfFormField clientSign = stamper.addSignature(fieldName, pos.page, pos.position.getLeft(), pos.position.getBottom(), pos.position.getRight(), pos.position.getTop());
					clientSign.setFlags(PdfAnnotation.FLAGS_PRINT);
				}
			}
		}
		stamper.close();
		copyOutputToTmp();
	}

	// Sign the PDF document :
	// - If 'fieldName' signature field given in params exits in the PDF, an appearance is made visible for this field, otherwise signature is applied without appearance.
	public void sign(SignParams params) throws Exception {
		boolean success = false;
		int maxTries = 20;
		for(int i=0; i<maxTries; i++) { // Use a limited loop to avoid infinite loop (just in case) !
			try {
				MyLogger.log("    Signing with estimatedSize=" + estimatedSignatureSize.get() + " bytes");
				sign_impl(params);
				success = true;
				break;
			} catch(IOException e) {
				if (e.getMessage().equals("Not enough space")) {
					// Our estimated size was too low, must retry with new estimated size
					MyLogger.log("    Not enough space for signature, retrying with estimatedSize=" + estimatedSignatureSize.get() + " bytes");
				} else {
					throw e;
				}
			}
		}
		if (!success) {
			throw new Exception("Enable to sign after " + maxTries + " tries");
		}
		copyOutputToTmp();
	}

	private void sign_impl(SignParams params) throws Exception {
		PdfReader reader = new PdfReader(new FileInputStream(tmpInput));
		boolean append = true;
		PdfStamper stamper = PdfStamper.createSignature(reader, new FileOutputStream(pdfOutput), '\0', null, append);
		PdfSignatureAppearance sap = stamper.getSignatureAppearance();

		if (params.fieldName != null) {
			boolean fieldVisible = reader.getAcroFields().getField(params.fieldName) != null;
			if (fieldVisible) {
				sap.setVisibleSignature(params.fieldName);

				if (params.fieldImage != null) {
					sap.setRenderingMode(PdfSignatureAppearance.RenderingMode.GRAPHIC);
					sap.setSignatureGraphic(params.fieldImage);
					sap.setLayer2Text("");
				} else {
					sap.setRenderingMode(PdfSignatureAppearance.RenderingMode.DESCRIPTION);
					sap.setLayer2Text("Signature électronique");
					sap.setLayer2Font(new Font(Font.FontFamily.COURIER, 6.0f));
				}
			}
		}

		sap.setCertificationLevel(PdfSignatureAppearance.NOT_CERTIFIED);
		Calendar cal = Calendar.getInstance();
		sap.setSignDate(cal);

		if (params.reason != null) {
			sap.setReason(params.reason);
		}
		if (params.location != null) {
			sap.setLocation(params.location);
		}
		if (params.contact != null) {
			sap.setContact(params.contact);
		}
		if (params.creator != null) {
			sap.setSignatureCreator(params.creator);
		}

		MyExternalSignatureContainer sigContainer = new MyExternalSignatureContainer(params.transactionId, params.signatureId);
		try {
			MakeSignature.signExternalContainer(sap, sigContainer, estimatedSignatureSize.get());
		} finally {
			synchronized (estimatedSignatureSize) {
				if (sigContainer.pkcs7Signature != null && estimatedSignatureSize.get() < sigContainer.pkcs7Signature.length) {
					estimatedSignatureSize.set(sigContainer.pkcs7Signature.length + 500); // Add some more bytes to ensure that we have enought space next time (signature size depends on lots of factors)
				}
			}
		}
		MyLogger.log("    Added signature to PDF");
	}

	// Timestamp the PDF document :
	// - timestamp contains a timestamp token and LTV information that will allow signature certificates verification when PDF is opened (under Adobe Reader for instance)
	// - LTV information includes : certificates OCSP responses and CRLs
	// - LTV information allows verifying signature certificates even if they have expired (without LTV information, Adobe Reader always displays a warning on signatures when
	//   their certificate has expired because, in that case, Certificate Authorities will not give certificate revocation state so Adobe Reader cannot verify that the certificate
	//   was not revoked at the date of signature)
	public void timestamp(String transactionId) throws Exception {
		boolean success = false;
		int maxTries = 20;
		for(int i=0; i<maxTries; i++) { // Use a limited loop to avoid infinite loop (just in case) !
			try {
				MyLogger.log("    Timestamping with estimatedSize=" + estimatedTimestampSize.get() + " bytes");
				timestamp_impl(transactionId);
				success = true;
				break;
			} catch(IOException e) {
				if (e.getMessage().equals("Not enough space")) {
					// Our estimated size was too low, must retry with new estimated size
					MyLogger.log("    Not enough space for timestamp, retrying with estimatedSize=" + estimatedTimestampSize.get() + " bytes");
				} else {
					throw e;
				}
			}
		}
		if (!success) {
			throw new Exception("Enable to sign after " + maxTries + " tries");
		}
		copyOutputToTmp();
	}

	private void timestamp_impl(String transactionId) throws Exception {
		PdfReader reader = null;
		Security.addProvider(new BouncyCastleProvider()); // Required for 'fields.verifySignature()' function, otherwise we get a null pointer exception

		try {
			reader = new PdfReader(new FileInputStream(tmpInput));
			PdfStamper stamper = PdfStamper.createSignature(reader, new FileOutputStream(pdfOutput), '\0', null, true);

			LtvVerification v = stamper.getLtvVerification();
			AcroFields fields = stamper.getAcroFields();
			List<String> sigNames = fields.getSignatureNames();
			String lastSigName = sigNames.get(sigNames.size() - 1);
			PdfPKCS7 lastSigPkcs7 = fields.verifySignature(lastSigName);

			OcspClientBouncyCastle ocspClient = new OcspClientBouncyCastle(new OCSPVerifier(null, null));
			CrlClientOnline crlClient = new CrlClientOnline();

			if (lastSigPkcs7.isTsp()) {
				// Add LVT info to the TimeStamp signature.
				v.addVerification(lastSigName, ocspClient, crlClient,
					LtvVerification.CertificateOption.WHOLE_CHAIN,
					LtvVerification.Level.OCSP_OPTIONAL_CRL, // OCSP alone does not work (in that case, signature is not LTV enabled).
					LtvVerification.CertificateInclusion.NO); // Seems useless to include certificates.

			} else {
				// Add LTV info for all signatures.
				for (String sigName : sigNames) {
					v.addVerification(sigName, ocspClient, crlClient,
						LtvVerification.CertificateOption.WHOLE_CHAIN, // WHOLE_CHAIN required for Adobe to display 'Signature is LTV enabled'.
						LtvVerification.Level.OCSP_OPTIONAL_CRL, // OCSP alone does not work (in that case, signature is not LTV enabled).
						LtvVerification.CertificateInclusion.NO); // Seems useless to include certificates.
				}
			}

			// Add the TimeStamp.
			MyTSAClient tsaClient = new MyTSAClient(transactionId);
			String tsSigName = toTsSigName(fields);

			try {
				LtvTimestamp.timestamp(stamper.getSignatureAppearance(), tsaClient, tsSigName);
			} finally {
				synchronized (estimatedTimestampSize) {
					if (tsaClient.timestamp != null && estimatedTimestampSize.get() < tsaClient.timestamp.length) {
						estimatedTimestampSize.set(tsaClient.timestamp.length + 200); // Add some more bytes to ensure that we have enought space next time (signature size depends on lots of factors)
					}
				}
			}
			MyLogger.log("    Added Timestamp to signatures");

		} finally {
			if (reader != null) {
				reader.close();
			}
		}
		copyOutputToTmp();
	}

	static private String toTsSigName(AcroFields fields) {
		String tsSigName;
		for(int i=1; i<=5; i++) {
			tsSigName = "SignatureLTV" + (i==1 ? "" : i);
			if (fields.getFieldItem(tsSigName) == null) {
				return tsSigName;
			}
		}
		// Should never reach this code.
		return "SignatureLTV_${System.currentTimeMillis()}";
	}

}
