package fr.contralia.client.example.test;

import com.itextpdf.text.Image;
import fr.contralia.client.example.Common;
import fr.contralia.client.example.PdfSigner;
import fr.contralia.client.example.XmlReader;
import static fr.contralia.client.example.MyLogger.*;

import java.io.*;

/**
 * Code sample calling Contralia API
 * 
 * @author lfaizant and fchristol
 *
 */
public class Main {

	// Example 1: sign a PDF locally using the signHashes and horodateHashes APIs
	private static void signTransaction() throws Exception {
		log("\n");
		log("----------------------------------");
		log("Sign transaction");
		log("----------------------------------");

		String result;
		XmlReader.Node2 xml;

		// Init transaction
		log("- Initializing transaction...");
		result = Common.initiate("ref", 1, false);
		xml = XmlReader.parse(result);
		String transactionId = xml.getAttribute("id");
		log(result + "\n");

		// Upload document to sign
		log("- Uploading document...");
		result = Common.upload(transactionId, "res/document.pdf", "Contract");
		log(result + "\n");

		// Set signatory
		log("- Setting signatory...");
		result = Common.signatory(transactionId, null, null, "0", "Truché", "Hervé", null, "0601020304", "Allée du Vallon",
			"Les Vaisseaux Argentés", "06901", "Sophia Antipolis", "France", null, "Chargé de mission", null, "PAD");
		xml = XmlReader.parse(result);
		String signatureId = xml.getAttribute("id");
		log(result + "\n");

		// Sign document
		log("- PAD signature...");
		result = Common.pad(signatureId, "res/signature.xml");
		log(result + "\n");

		// Terminate transaction
		log("- Terminating transaction ...");
		result = Common.terminate(transactionId);
		log(result  + "\n");

		// Download signed document
		log("- Downloading signed document...");
		File signedDoc = new File("target", "signTransaction_finalDoc.pdf");
		Common.finalDoc(transactionId, "Contract", signedDoc);
		log("Signed doc created in '" + signedDoc.getAbsolutePath() + "'\n");
	}

	// Example 2 : sign a PDF locally using the signHashes and horodateHashes APIs
	private static void signPdfLocally() throws Exception {
		log("\n");
		log("----------------------------------");
		log("Sign PDF locally");
		log("----------------------------------");

		// Open document to sign
		log("- Opening PDF document...");
		File pdfFile = new File("res/document.pdf");
		File signedPdfFile = new File("target", "signPdfLocally_finalDoc.pdf");
		File tmpPdfFile = new File("target", "tmp.pdf"); // tmp file is required to store temporary files during signature & timestamp operations

		String result;
		XmlReader.Node2 xml;

		// Init transaction
		log("- Initializing transaction...");
		result = Common.initiate("ref", 1, false);
		xml = XmlReader.parse(result);
		String transactionId = xml.getAttribute("id");
		log(result + "\n");

		// Set signatory
		log("- Setting signatory...");
		result = Common.signatory(transactionId, null, null, "0", "Truché", "Hervé", null, "0601020304", "Route de DOCAPOSTE",
			"Batiment A", "06901", "Sophia Antipolis", "France", null, "Chargé de mission", null, "PAD");
		xml = XmlReader.parse(result);
		String signatureId = xml.getAttribute("id");
		log(result + "\n");

		// Sign PDF document with signHashes API
		PdfSigner pdfsigner = new PdfSigner(pdfFile, signedPdfFile, tmpPdfFile);

		// 1/ Replace text fields by signature fields
		pdfsigner.replaceFields();

		// 2/ Perform signatory signature
		PdfSigner.SignParams params = new PdfSigner.SignParams();
		params.transactionId = transactionId;
		params.signatureId = signatureId;
		params.fieldName = "SignatureContractant";
		params.fieldImage = Image.getInstance("res/signatorySignature.png"); // Note : we used a simple image, but you may want to prepare a more complex image containing signatory name and date
		params.reason = "My signatory reason";
		params.location = "My signatory location";
		params.contact = "My signatory contact";
		params.creator = "My signatory creator";
		pdfsigner.sign(params);

		// 3/ Perform organisation signature
		params = new PdfSigner.SignParams();
		params.transactionId = transactionId;
		params.signatureId = "1"; // by convention (see signHashes API doc)
		params.fieldName = "SignatureFournisseur";
		params.fieldImage = Image.getInstance("res/organizationSignature.png"); // Note : we used a simple image, but you may want to prepare a more complex image containing signatory name and date
		params.reason = "My organization reason";
		params.location = "My organization location";
		params.contact = "My organization contact";
		params.creator = "My organization creator";
		pdfsigner.sign(params);

		// 3/ Perform Contralia signature
		params = new PdfSigner.SignParams();
		params.transactionId = transactionId;
		params.signatureId = "2"; // by convention (see signHashes API doc)
		params.fieldName = "TamponContralia";
		params.fieldImage = Image.getInstance("res/sealingSignature.png"); // Note : we used a simple image, but you may want to prepare a more complex image containing signatory name and date
		params.reason = "My Contralia reason";
		params.location = "My Contralia location";
		params.contact = "My Contralia contact";
		params.creator = "My Contralia creator";
		pdfsigner.sign(params);

		// Horodate PDF document with horodateHashes API
		// 2/ Perform signatory signature
		pdfsigner.timestamp(transactionId);

		// Terminate transaction
		log("Terminating transaction ...");
		result = Common.terminate(transactionId);
		log(result + "\n");

		log("Signed doc created in '" + signedPdfFile.getAbsolutePath() + "'\n");
	}

	public static void main(String[] args) {

		try {
			String testName = args.length > 0 ? args[0] : null;

			if (testName == null || testName.equals("signTransaction")) {
				signTransaction();
			}

			if (testName == null || testName.equals("signPdfLocally")) {
				signPdfLocally();
			}

		} catch(Exception e) {

			// Search for a HttpException which is the real source of the error
			Common.HttpException he = null;
			if (e.getCause() instanceof Common.HttpException) {
				he = (Common.HttpException) (e.getCause());
			} else if (e.getCause() != null && e.getCause().getCause() instanceof Common.HttpException) {
				he = (Common.HttpException) (e.getCause().getCause());
			}

			if (he != null) {
				logError("statusCode=" + he.statusCode + ", reasonPhrase=" + he.reasonPhrase + ", response=" + he.response);
			} else {
				e.printStackTrace();
			}
		}
	}

}
